{ *****************************************************************************
www:          http://bitbandit.org/
Author(s):    Lajos Bicsk (asimo@bitbandit.org)

product:      z88cpy
Version:      1.0

Description:  Cambridge Z88 file transfer tool for Windows.

Legal issues: Copyright (C) 2014 Lajos Bicsk

              This software is provided 'as-is', without any express or implied
              warranty.  In no event will the author be held liable for any
              damages arising from the use of this software.

              Permission is granted to anyone to use this software for any
              purpose, including commercial applications, and to alter it and 
              redistribute it freely, subject to the following restrictions:

              1. The origin of this software must not be misrepresented; you
                 must not claim that you wrote the original software. If you use
                 this software in a product, an acknowledgment in the product
                 documentation would be appreciated but is not required.

              2. Altered source versions must be plainly marked as such, and must
                 not be misrepresented as being the original software.

              3. This notice may not be removed or altered from any source
                 distribution.

History:      04/10/2014  Initial 1.0 release

****************************************************************************** }
{$MODE DELPHI}
{$R *.res}
program z88cpy;

// Hard reset the Z88: open card flap, press reset switch twice, shut card flap back

// Cambridge Z88 Import-Export file transfer protocol:
// ESC, N, filename, ESC, F, file, ESC, E (or ESC Z if batch transfer)
// ESC B x y - binary data x,y nibbles ie. transfer A9 => ESC,B,A,9

uses
  Windows, SysUtils;

const
  BoolStr: Array[False..True] of String = ('Disabled', 'Enabled');
  ESC = $1B;

type
  AoB = Array of Byte;

var
  hCom: Integer = -1;
  PortError: Boolean;
  ConIn, ConOut: THandle;

function HexToByte(A, B: Byte): Byte;
begin
  Dec(A, 48);
  Dec(B, 48);
  if (A > 9) then
    Dec(A, 7);
  if (B > 9) then
    Dec(B, 7);
  Result := A shl 4 or B;
end;

function ToHex(Nibble: Byte): Byte;
begin
  Result := Nibble + 48;
  if (Result > 57) then
    Inc(Result, 7);
end;

procedure MoveBackCursor;
var
  Info: TConsoleScreenBufferInfo;
begin
  if GetConsoleScreenBufferInfo(ConOut, Info) then
  begin
    Info.dwCursorPosition.X := 0;
    SetConsoleCursorPosition(ConOut, Info.dwCursorPosition);
  end;
end;

function IsEscPressed: Boolean;
var
  InRec: TInputRecord;
  Cnt: Cardinal;
begin
  Result := False;
  if PeekConsoleInput(ConIn, @InRec, 1, @Cnt) then
    if (Cnt = 1) then
    begin
      ReadConsoleInput(ConIn, @InRec, 1, @Cnt);
      Result := (InRec.EventType = KEY_EVENT) and
                (InRec.Event.KeyEvent.bKeyDown) and
                (InRec.Event.KeyEvent.wVirtualKeyCode = VK_ESCAPE);
    end;
end;


function OpenComPort(Port: String; Baud: Integer; XonXoff: Boolean): Boolean;
const
  ComTimeouts: TCOMMTIMEOUTS = (
    ReadIntervalTimeout         : 1000;
    ReadTotalTimeoutMultiplier  : 1000;
    ReadTotalTimeoutConstant    : 0;
    WriteTotalTimeoutMultiplier : 10000;
    WriteTotalTimeoutConstant   : 0;
  );
var
  APort: String;
  ComState: TDCB;
begin
  APort := '\\.\' + Port;
  hCom := CreateFile(
    PChar(APort),
    GENERIC_READ or GENERIC_WRITE,
    0,
    nil,
    OPEN_EXISTING,
    FILE_ATTRIBUTE_NORMAL,
    0
  );
  Result := (hCom <> -1);
  if Result then
  begin
    SetCommTimeouts(hCom, ComTimeouts);
    GetCommState(hcom, ComState);
    ComState.BaudRate := Baud;
    ComState.ByteSize := 8;
    ComState.StopBits := 0; // 0 == 1 stop bit
    ComState.flags := $6021;
    if XonXoff then
      ComState.flags := ComState.flags or $0300;
// 0000 0000 0000 0000 0110 0011 0010 0001
// 00. fBinary = 1
// 01. fParity = 0
// 02. fOutxCtsFlow = 0
// 03. fOutxDsrFlow = 0
// 04. fDtrControl = Handshake (10)
// 06. fDsrSensitivity = 0
// 07. fTXContinueOnnXoff = 0
// 08. fOutX = ?
// 09. fInX = ?
// 10. fErrorChar = 0
// 11. fNull = 0
// 12. fRtsControl = Handshake (10)
// 14. fAbortOnError = 1
    SetCommState(hcom, ComState);
  end;
end;

procedure CloseComPort;
begin
  if (hCom <> -1) then
  begin
    CloseHandle(hCom);
    hCom := -1;
  end;
end;

procedure SendByte(AByte: Byte);
var
  BytesSent: Cardinal;
begin
  PortError := True;
  if (hCom <> -1) then
    if WriteFile(hCom, AByte, 1, BytesSent, nil) then
      PortError := False
    else
      WriteLn(Format('Error during send: %d', [GetLastError]));
end;

function SendBytes(Buffer: AoB; Len: Cardinal): Cardinal;
begin
  Result := 0;
  PortError := True;
  if (hCom <> -1) then
    if WriteFile(hCom, Buffer[0], Len, Result, nil) then
      PortError := False
    else
      WriteLn(Format('Error during send: %d', [GetLastError]));
end;

function ReadByte: Byte;
var
  BytesRead: Cardinal;
begin
  result := 0;
  PortError := True;
  if (hCom <> -1) then
    if ReadFile(hCom, Result, 1, BytesRead, nil) then
      PortError := False
    else
      WriteLn(Format('Error during receive: %d', [GetLastError]));
end;

// 0=OK, 1=Timeout/Abort, 2=Error
function ReceiveFile(var AFilename: string; var AOut: AoB; var EndOfBatch: Boolean; Timeout: Integer): Integer;
var
  I: Integer;
  Filename: string;
  A, B: Byte;
begin
  Result := 2;
  EndOfBatch := False;

  // Waiting for the first ESC,N sequence (max. 15 seconds)
  // The buffer may contain other bytes come from previously interrupted transfer
  I := Timeout;
  while (I > 0) do
  begin
    if IsEscPressed then
    begin
      WriteLn('Esc key pressed, abort');
      Result := 1;
      Exit;
    end;
    Write(Format('Waiting... %d   ', [I]));
    MoveBackCursor;
    Dec(I);
    B := ReadByte;
    if PortError then
      Exit;
    if (B <> 27) then
      Continue;
    B := ReadByte;
    if PortError then
      Exit;
    if (B = Ord('N')) then
      Break;
  end;
  if (I <= 0) then
  begin
    WriteLn('Receive timeout.               ');
    Result := 1;
    Exit;
  end;

  I := 0;
  SetLength(Filename, 256);
  while I <= 256 do
  begin
    B := ReadByte;
    if (B in [0, 27]) then
      Break;
    Inc(I);
    Filename[I] := Char(B);
  end;
  if (B <> 27) then
    Exit;
  SetLength(Filename, I);
  WriteLn(Format('Receiving file: %s', [Filename]));
  AFilename := Filename;

  if (ReadByte <> Ord('F')) then
    Exit;
  I := 0;
  Write('Received bytes: 0');
  MoveBackCursor;
  SetLength(AOut, 16384);
  while True do
  begin
    B := ReadByte;
    if (B = 27) then
    begin
      B := ReadByte;
      case Char(B) of
        'Z':  // not supported by Z88's Import/Export function
        begin
          EndOfBatch := True;
          Result := 0;
          Break;
        end;
        'E':  // multiple file transfer 
        begin
          EndOfBatch := False;
          Result := 0;
          Break;
        end;
        'B':
        begin
          A := ReadByte;
          B := ReadByte;
          if not (Char(A) in ['0'..'9', 'A'..'F']) then
            Break;
          if not (Char(B) in ['0'..'9', 'A'..'F']) then
            Break;
          B := HexToByte(A, B);
        end;
        else
          Break;
      end;
    end else
    if not (B in [9,10,13]) and ((B < $20) or (B > $7E)) then
      Break;
    if (I >= Length(AOut)) then
      SetLength(AOut, Length(AOut) + 16384);
    AOut[I] := B;
    Inc(I);
    if (I mod 10 = 0) then
    begin
      Write(Format('Received bytes: %d', [I]));
      MoveBackCursor;
    end;
  end;
  WriteLn(Format('Received bytes: %d', [I]));
  SetLength(AOut, I);  
end;

function SendFile(const AIn: AoB; AFilename: String; EndOfBatch: Boolean): Integer;
var
  Header: AoB;
  HLen: Integer;
  Hex: AoB;
  AByte: Byte;
  I: Integer;
begin
  Result := 0;
  HLen := Length(AFilename) + 4;
  SetLength(Header, HLen);
  Header[0] := ESC;
  Header[1] := Ord('N');
  Move(AFilename[1], Header[2], Length(AFilename));
  Header[HLen-2] := ESC;
  Header[HLen-1] := Ord('F');
  SendBytes(Header, HLen);
  if PortError then
    Exit;

  SetLength(Hex, 4);
  Hex[0] := ESC;
  Hex[1] := Ord('B');
  Write('Bytes sent: 0');
  MoveBackCursor;
  for I := 0 to Length(AIn)-1 do
  begin
    AByte := AIn[I];
    if not (AByte in [9,10,13]) and ((AByte < $20) or (AByte > $7E)) then
    begin
      Hex[2] := ToHex(AByte div 16); 
      Hex[3] := ToHex(AByte mod 16); 
      SendBytes(Hex, 4);
    end else
      SendByte(AByte);
    if PortError then
      Exit;
    Inc(Result);
    if (I mod 10 = 0) then
    begin
      Write(Format('Bytes sent: %d', [Result]));
      MoveBackCursor;
    end;
  end;
  WriteLn(Format('Bytes sent: %d', [Result]));

  SendByte(ESC);
  if PortError then
    Exit;
  if EndOfBatch then
    SendByte(Ord('Z'))
  else
    SendByte(Ord('E'));
end;

type
  TransferMode = (TM_NONE, TM_SEND, TM_RECEIVE);

var
  F: File;
  InBuf: AoB;
  Len, Sent: Integer;
  Files: Array of String;
  FileCnt: Integer;
  SendFiles: Array of String;
  SendFileCnt: Integer;
  SentCnt: Integer;
  ComNum, Baud: Integer;
  I: Integer;
  XonXoff: Integer;
  Param: String;
  Mode: TransferMode;
  FilePath: String;
  FullFilename: String;
  Filename: String;
  EoT: Boolean;
  FindData: TWin32FindData;
  FHandle: THandle;
  Timeout: Integer;
  RecvRes: Integer;

begin
  ConOut := GetStdHandle(STD_OUTPUT_HANDLE);
  ConIn := GetStdHandle(STD_INPUT_HANDLE);

  if ParamCount > 0 then
  begin
    SetLength(Files, 0);
    FileCnt := 0;
    ComNum := 1;
    Mode := TM_NONE;
    Baud := 9600;
    XonXoff := 1;
    Timeout := 15;
    for I:=1 to ParamCount do
    begin
      Param := ParamStr(I);
      if Param[1] = '-' then
      begin
        Delete(Param, 1, 1);
        if Length(Param) < 1 then
          Continue;
        case Param[1] of
          's', 'S': Mode := TM_SEND;
          'r', 'R': Mode := TM_RECEIVE;
          'x', 'X':
          begin
            Delete(Param, 1, 1);
            XonXoff := StrToIntDef(Param, -1);
            if not (XonXoff in [0, 1]) then
            begin
              WriteLn('Invalid Xon/Xoff mode.');
              Exit;
            end;
          end;
          'b', 'B':
          begin
            Delete(Param, 1, 1);
            Baud := StrToIntDef(Param, 0);
            if (Baud <> 9600) and (Baud <> 19200) and (Baud <> 38400) then
            begin
              WriteLn(Format('Invalid baud rate: %d', [Baud]));
              Exit;
            end;
          end;
          'p', 'P':
          begin
            Delete(Param, 1, 1);
            ComNum := StrToIntDef(Param, 0);
            if (ComNum <= 0) then
            begin
              WriteLn('Invalid COM number.');
              Exit;
            end;
          end;
          't', 'T':
          begin
            Delete(Param, 1, 1);
            Timeout := StrToIntDef(Param, 0);
            if (Timeout <= 0) then
            begin
              WriteLn('Invalid timeout value.');
              Exit;
            end;
          end;
        end;
      end else
      begin
        SetLength(Files, FileCnt+1);
        Files[FileCnt] := Param;
        Inc(FileCnt);
      end;
    end;

    if (Mode = TM_NONE) then
    begin
      WriteLn('Please provide send or receive mode.');
      Exit;
    end;

    if (Mode = TM_SEND) and (FileCnt = 0) then
    begin
      WriteLn('Please give at least one filename.');
      Exit;
    end;

    WriteLn(Format('Selected port: COM%d, Baud rate: %d, Xon/Xoff: %s', [ComNum, Baud, BoolStr[Boolean(XonXoff)]]));

    if OpenComPort(Format('COM%d', [ComNum]), Baud, Boolean(XonXoff)) then
    begin
      WriteLn('COM port opened.');

      if (Mode = TM_SEND) then
      begin
        SentCnt := 0;
        SendFileCnt := 0;
        for I := 0 to FileCnt-1 do
        begin
          FilePath := ExtractFilePath(Files[I]);
          FHandle := FindFirstFile(PChar(Files[I]), FindData);
          if (FHandle <> INVALID_HANDLE_VALUE) then
          begin
            while (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0) do
            begin
              SetLength(SendFiles, SendFileCnt+1);
              SendFiles[SendFileCnt] := FilePath + StrPas(@FindData.cFilename[0]);
              Inc(SendFileCnt);
              if not FindNextFile(FHandle, FindData) then
                Break;
            end;
            Windows.FindClose(FHandle);
          end;
        end;
        for I := 0 to SendFileCnt-1 do
        begin
          FullFilename := SendFiles[I];
          Filename := ExtractFilename(FullFilename);
          WriteLn(Format('Sending file: %s', [Filename]));
          FileMode := fmOpenRead;
          AssignFile(F, FullFilename);
          try
            try
              Reset(F, 1);
              Len := FileSize(F);
              SetLength(InBuf, Len);
              BlockRead(F, InBuf[0], Len);
            finally
              CloseFile(F);
            end;
          except
            on E: Exception do
            begin
              WriteLn(Format('Error while reading file: %s', [E.Message]));
              Break;
            end;
          end;
          WriteLn(Format('File read, length: %d', [Len]));
          Sent := SendFile(InBuf, Filename, I = SendFileCnt-1);
          if PortError or (Sent <> Len) then
            Break;
          Inc(SentCnt);
        end;
        if (SentCnt = 0) then
          WriteLn('No file has been sent.')
        else if (SentCnt = 1) then
          WriteLn('One file has been sent successfully.')
        else
          WriteLn(Format('%d files have been sent successfully.', [SentCnt]));
      end;

      if (Mode = TM_RECEIVE) then
      begin
        while True do
        begin
          RecvRes := ReceiveFile(Filename, InBuf, EoT, Timeout);
          if RecvRes > 0 then
          begin
            if RecvRes > 1 then
              WriteLn('Error while receiving data.');
            Break;
          end;
          Len := Length(InBuf);
          Filename := ExtractFilename(Filename);
          FileMode := fmOpenReadWrite;
          AssignFile(F, Filename);
          try
            try
              Rewrite(F, 1);
              BlockWrite(F, InBuf[0], Len);
            finally
              CloseFile(F);
            end;
          except
            on E: Exception do
            begin
              WriteLn(Format('Error while writing file: %s', [E.Message]));
              Break;
            end;
          end;
          WriteLn(Format('File saved as %s', [Filename]));
          if EoT then
          begin
            WriteLn('All files have been received.');
            Break;
          end;
        end;
      end;

      CloseComPort;
      WriteLn('COM port closed.');
    end else
      WriteLn('Failed to open COM port.');
  end else
  begin
    WriteLn('Z88 Copy version 1.0 [2014/10/04]');
    WriteLn('Copyright (c) 2014 by Lajos Bicsak. This tool is freeware. No warranty.');
    WriteLn('Transfers files between PC and Z88 through the serial port using Import/Export.');
    WriteLn('COM port settings: 8 data-bit, 1 stop-bit, no parity.');
    WriteLn('Usage: z88cpy [options] [file1 file2 .. fileN]');
    WriteLn('Options:');
    WriteLn('  -r     to receive file');
    WriteLn('  -s     to send file(s), wildcards can be used');
    WriteLn('  -pNUM  where NUM is COM port number, default is 1 for COM1');
    WriteLn('  -bNUM  where NUM is baud rate: 9600, 19200, 38400, default is 9600');
    WriteLn('  -xNUM  where NUM=0 to disable or NUM=1 to enable Xon/Xoff, default is enabled');
    WriteLn('  -tNUM  where NUM is the wait timeout in seconds while receiving, default is 15');
    WriteLn('Example:');
    WriteLn('  z88cpy -s -p2 -b19200 -x0 jsw.* installer.bas');
    WriteLn('Notes:');
    WriteLn('  To start receiving on Z88 select Import/Export in the Index menu and press B.');
    WriteLn('  To start sending on Z88 press S and enter filename in Import/Export.');
    WriteLn('  Z88''s Import/Export function does not indicate the end of batch sending, ie.');
    WriteLn('  sending multiple files by wildcards. Therefore receiving must be terminated');
    WriteLn('  by pressing the ESC key on the PC side when finished, or wait until timeout.');
  end;

end.
